/*
 * Copyright (c) 1993,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).  All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 *      Utah $Hdr: act_subr.c 1.4 95/01/26$
 */
/*
 * Support routines for the migrating thread path.
 *
 * These are called on a kernel activation (i.e. are not glue code)
 * from glue code.
 */
#include <mach/port.h>
#include <mach/message.h>
#include <kern/lock.h>
#include <kern/task.h>
#include <kern/current.h>
#include <ipc/ipc_kmsg.h>
#include <ipc/ipc_mqueue.h>
#include <ipc/ipc_object.h>
#include <ipc/ipc_port.h>
#include <ipc/ipc_space.h>
#include <ipc/ipc_entry.h>
#include <ipc/mach_msg.h>
#include <mach/rpc.h>

/*
 *
 */


/*
 * Special routines for migrating thread RPC.
 *
 * If the request/reply part of an RPC is determined to be COMPLEX,
 * these routines are called to properly handle copying port rights
 * and out of line data from a server message to the client or visa-versa.
 *
 * `smsg' is the client message which comes from the task `stask'.
 * `dmsg' is the server message which goes to the task `dtask'.
 */

mach_msg_return_t
rpc_copy_request(smsg, dmsg, size, stask, dtask)
	mach_msg_header_t *smsg;
	mach_msg_header_t *dmsg;
	mach_msg_size_t size;
	task_t stask;
	task_t dtask;
{
	ipc_kmsg_t kmsg;
	mach_msg_return_t mr;

	/*
	 * XXX current activation for copyinmsg set to client.
	 */
	cur_pcb->ss.act->mact.user_act = cur_pcb->ss.act->lower->lower;

	mr = ipc_kmsg_get(smsg, size, &kmsg);
	if (mr != MACH_MSG_SUCCESS)
		return mr;

	/*
	 * XXX ugh!  Destination (remote) port is still in the context of
	 * the source (client).  Convert to internal form.
	 */
	{
		ipc_entry_t entry;

		is_read_lock(stask->itk_space);
		entry = ipc_entry_lookup(stask->itk_space,
					 kmsg->ikm_header.msgh_remote_port);
		is_read_unlock(stask->itk_space);
		if (entry == IE_NULL) {
			ikm_free(kmsg);
			return MACH_SEND_INVALID_DEST;
		}
		kmsg->ikm_header.msgh_remote_port =
			(mach_port_t) entry->ie_object;
	}

	mr = ipc_kmsg_copyin_body(kmsg, stask->itk_space, stask->map);
	if (mr != MACH_MSG_SUCCESS) {
		ikm_free(kmsg);
		return mr;
	}

	/*
	 * XXX current activation for copyoutmsg set to server.
	 */
	cur_pcb->ss.act->mact.user_act = cur_pcb->ss.act->lower;

	if (kmsg->ikm_header.msgh_bits & MACH_MSGH_BITS_COMPLEX) {
		vm_offset_t saddr, eaddr;

		saddr = (vm_offset_t) (&kmsg->ikm_header + 1);
		eaddr = (vm_offset_t) &kmsg->ikm_header +
				kmsg->ikm_header.msgh_size;

		mr = ipc_kmsg_copyout_body(saddr, eaddr,
					   dtask->itk_space, dtask->map);
		if (mr != MACH_MSG_SUCCESS)
			mr |= MACH_RCV_BODY_ERROR;
	}
	if (mr != MACH_MSG_SUCCESS) {
		/* XXX what to do on failure? */
		;
	}

	/*
	 * XXX maintain the sequence number
	 */
	{
		ipc_port_t port;

		port = (ipc_port_t) kmsg->ikm_header.msgh_remote_port;
		kmsg->ikm_header.msgh_seqno = port->ip_seqno++;
	}

	/*
	 * XXX swap the local and remote ports and translate the destination
	 * port name to the client's space.
	 */
	{
		mach_port_t tmp;

		tmp = kmsg->ikm_header.msgh_remote_port;
		kmsg->ikm_header.msgh_remote_port = 
			kmsg->ikm_header.msgh_local_port;
		kmsg->ikm_header.msgh_local_port =
			((ipc_port_t)tmp)->ip_receiver_name;
	}

	mr = ipc_kmsg_put(dmsg, kmsg, size);
	return mr;
}

mach_msg_return_t
rpc_copy_reply(smsg, stask, dmsg, dtask, size)
	mach_msg_header_t *smsg;
	task_t stask;
	mach_msg_header_t *dmsg;
	task_t dtask;
	mach_msg_size_t size;
{
	ipc_kmsg_t kmsg;
	mach_msg_return_t mr;

	/*
	 * XXX current activation for copyinmsg set to server.
	 */
	cur_pcb->ss.act->mact.user_act = cur_pcb->ss.act->lower;

	mr = ipc_kmsg_get(smsg, size, &kmsg);
	if (mr != MACH_MSG_SUCCESS)
		return mr;

	/*
	 * XXX ugh!  Destination (reply) port is still in the context of
	 * the destination (client).  Convert to internal form.
	 */
	{
		ipc_entry_t entry;

		is_read_lock(dtask->itk_space);
		entry = ipc_entry_lookup(dtask->itk_space,
					 kmsg->ikm_header.msgh_remote_port);
		is_read_unlock(dtask->itk_space);
		if (entry == IE_NULL) {
			ikm_free(kmsg);
			return MACH_SEND_INVALID_DEST;
		}
		kmsg->ikm_header.msgh_remote_port =
			(mach_port_t) entry->ie_object;
	}

	mr = ipc_kmsg_copyin_body(kmsg, stask->itk_space, stask->map);
	if (mr != MACH_MSG_SUCCESS) {
		ikm_free(kmsg);
		return mr;
	}

	/*
	 * XXX current activation for copyoutmsg set to client.
	 */
	cur_pcb->ss.act->mact.user_act = cur_pcb->ss.act->lower->lower;

	if (kmsg->ikm_header.msgh_bits & MACH_MSGH_BITS_COMPLEX) {
		vm_offset_t saddr, eaddr;

		saddr = (vm_offset_t) (&kmsg->ikm_header + 1);
		eaddr = (vm_offset_t) &kmsg->ikm_header +
				kmsg->ikm_header.msgh_size;

		mr = ipc_kmsg_copyout_body(saddr, eaddr,
					   dtask->itk_space, dtask->map);
		if (mr != MACH_MSG_SUCCESS)
			mr |= MACH_RCV_BODY_ERROR;
	}
	if (mr != MACH_MSG_SUCCESS) {
		/* XXX what to do on failure? */
		;
	}

	/* XXX need to change any header values? seqno? size? */

	mr = ipc_kmsg_put(dmsg, kmsg, size);
	return mr;
}
