/* 
 * Copyright (c) 1990, 1991, 1994 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software and its
 * documentation is hereby granted, provided that both the copyright
 * notice and this permission notice appear in all copies of the
 * software, derivative works or modified versions, and any portions
 * thereof, and that both notices appear in supporting documentation.
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: mux.c 1.8 94/11/10$
 * 	Author: Leigh Stoller, Bob Wheeler, University of Utah CSL
 */

/*
 *	Driver for CIO 6 port serial mux card.
 */

#include <mux.h>

#if NMUX > 0

/*
 * UNIX wants to make termials 7 bit 1 stop no parity but the tip lines are
 * always in raw mode (8 bits 1 stop no parity). If TIP_DEBUGGING is defined
 * then all terminals are set to 8 bits 1 stop no parity and any attempt to 
 * change them is ignored.
 */
#define TIP_DEBUGGING

#include <sys/types.h>
#include <device/tty.h>
#include <device/errno.h>
#include <device/io_req.h>
#include <machine/spl.h>

#include <hpcio/device.h>
#include <hpcio/cioreg.h>
#include <hpcio/muxreg.h>
#include <hpcio/z80.h>

struct mux_softc {
	struct cio_chan    *cc;		/* pointer to cio channel struct */
	struct cio_request *cr;		/* cio request structure */
	u_char		   *inbuf;	/* Buffer for incoming data */
	u_char		   *outbuf;	/* Buffer for outgoing data */
	u_char		   *bp;		/* outbuf pointer */
	struct dmaquad     *cmdq;	/* general command quad */
	struct dmaquad     *write;	/* quad for reading from mux */
	struct dmaquad     *read;	/* quad for writing to mux */
	struct linkquad    *link;	/* link quad for completion */
	struct semaquad    *sema;	/* semaphore quad */
	struct statquad    *stat;	/* status quad */
	struct dmaquad     *quads;	/* quads for scatter/gather */
	int		   alive;	/* mux is alive and configured? */
	u_int		   soft_carrier;/* least significant bit is port 0 */
	int		   counts[6];	/* Space left in transmit buffer */
	u_int		   modem[6];    /* modem state for port */
	struct	tty	   tty[6];	/* tty structures for this mux */
};

int	mux_polled = 0;			/* muxes are being polled */

struct	mux_softc mux_softc[NMUX];
struct	cio_request muxcr[NMUX];

#if	KGDB
extern int kgdb_dev;
#endif	KGDB

#define	BOARD(x)	(((x) >> 3) & 0x1f)
#define PORT(x)		((x) & 7)

#define muxpollon()	timeout(mux_poll, 69, 4)
#define muxpolloff()	untimeout(mux_poll, 69)

int	mux_poll(), muxstart(), muxstop();
void	muxintr();

int	muxattach(), muxprobe(), muxscan(), muxlayer();
void	muxdisplay();
struct driver muxdriver = {
	"mux", muxattach, muxprobe, muxscan, muxdisplay, muxlayer
};

muxprobe(cc)
	struct cio_chan *cc;
{
	return (cc->type == CIOIDY_MUX) ? 0 : ENXIO;
}

muxattach(cc, unit, flags)
	struct cio_chan *cc;
{
	register struct mux_softc *ms = &mux_softc[unit];
	struct dmaquad *q;
	register u_char *bp;
	int port;

	if (cc->alive)
		return 0;

	if (unit > NMUX)
		return ENXIO;

	/*
	 * save cio channel number for this mux
	 */
	ms->cc = cc;
	ms->soft_carrier = flags;

	/*
	 * allocate buffers to store characters for input and output
	 */
	ms->inbuf  = (u_char *) cio_malloc(MUX_OUT_BUFSIZE, CIO_BUFFER_ALIGN);
	ms->outbuf = (u_char *) cio_malloc(MUX_IN_BUFSIZE, CIO_BUFFER_ALIGN);
	ms->bp = ms->outbuf;

	/*
	 * allocate the dma quads that we will need
	 */
	q = (struct dmaquad *) cio_malloc(sizeof(struct dmaquad) * 11,
					 CIO_QUAD_ALIGN);

	ms->cmdq  = q++;
	ms->write = q++;
	ms->read  = q++;
	ms->link  = (struct linkquad *) q++;
	ms->stat  = (struct statquad *) q++;
	ms->sema  = (struct semaquad *) q++;
	ms->quads = q;

	ms->cmdq->link    = (struct dmaquad *) ms->link;
	ms->cmdq->address = (caddr_t) ms->outbuf;

	ms->write->link    = (struct dmaquad *) ms->link;
	ms->write->command = CIODMA_WD | CIO_BYTE;
	ms->write->address = (caddr_t) ms->outbuf;

	ms->read->link    = (struct dmaquad *) ms->link;
	ms->read->command = CIODMA_RD | CIO_BYTE;
	ms->read->address = (caddr_t) ms->inbuf;
	ms->read->count   = MUX_IN_BUFSIZE;

	ms->link->link    = CIO_EOC(ms->link);
	ms->link->command = CIODMA_LNKST;
	ms->link->sema    = ms->sema;
	ms->link->stat    = ms->stat;

	/*
	 * set up the CIO request information
	 */
	ms->cr		  = &muxcr[unit];
	ms->cr->cmdq 	  = ms->cmdq;
	ms->cr->stat 	  = ms->stat;
	ms->cr->quad      = ms->quads;
	ms->cr->ciounit	  = cc->ciounit;
	ms->cr->channel   = cc->channel;
	ms->cr->polled    = 0;

	/* 
	 * Initialize the Mux for downloading. 
	 */
	ms->cmdq->command = CIODMA_DLD | CIO_CODE;
	bp = ms->outbuf;
	mux_initialize_download(bp);
	ms->cmdq->count = bp - ms->outbuf;
	if (ciosync(ms->cr)) {
		printf("failed to initialize download for mux%d\n",
		       unit);
		return EIO;
	}

	/* 
	 * Set the ram address for the download data 
	 */
	ms->cmdq->command = CIODMA_DLD | CIO_CODE;
	bp = ms->outbuf;
	mux_set_download_address(bp, Z80ADD, Z80SIZE);
	ms->cmdq->count = bp - ms->outbuf;
	if (ciosync(ms->cr)) {
		printf("failed to set download ram address on mux%d\n",
		       unit);
		return EIO;
	}

	/* 
	 * Now send the z80 code in one big DMA operation. 
	 */
	ms->cmdq->command = CIODMA_DLD;
	ms->cmdq->count   = Z80SIZE;
	bcopy(z80code, ms->outbuf, Z80SIZE);
	if (ciosync(ms->cr)) {
		printf("failed to download code to mux%d\n", unit);
		return EIO;
	}

	/* 
	 * Send a verify download command. 
	 */
	ms->cmdq->command = CIODMA_DLD | CIO_CODE;
	bp = ms->outbuf;
	mux_verify_download(bp);
	ms->cmdq->count = bp - ms->outbuf;
	if (ciosync(ms->cr)) {
		printf("failed to verify download on mux%d\n", unit);
		return EIO;
	}

	/* 
	 * Read back status (and ignore it)
	 */
	ms->cmdq->command = CIODMA_RDS | CIO_CODE;
	bp = ms->inbuf;
	ms->cmdq->count = CIO_MIN_BUFFER;
	if (ciosync(ms->cr)) {
		printf("failed to read status on mux%d\n", unit);
		return EIO;
	}

	/* 
	 * Send a start command to the mux
	 */
	ms->cmdq->command = CIODMA_DLD | CIO_CODE;
	bp = ms->outbuf;
	mux_start_execution(bp, Z80ADD);
	ms->cmdq->count = bp - ms->outbuf;
	if (ciosync(ms->cr)) {
		printf("failed to start mux%d\n", unit);
		return EIO;
	}

	/*
	 * set all the ports to the following default configuration
	 *
	 * modem transmittion mode : full duplex hardwired
	 * baud rate		   : 9600
	 * hardware XON/XOFF 	   : disabled
	 * char len		   : 7 bits, 1 stop, no parity
	 * modem connection timer  : 25 seconds
	 * no activity timer       : 0 seconds
	 * lost receiver timer	   : 400 msec
	 * receiver state	   : off
	 */
	ms->cmdq->command = CIODMA_WD | CIO_BYTE;
	bp = ms->outbuf;
	for (port = 0; port < 6; port++) {
		mux_baudrate(bp, port, BAUD_9600);
#ifdef TIP_DEBUGGING
		mux_char_length(bp, port, BITS_8 | STOP_1 | PARITY_NONE);
#else
		mux_char_length(bp, port, BITS_7 | STOP_1 | PARITY_NONE);
#endif
		mux_modem_connect_timer(bp, port, 25);
		mux_lost_receiver_timer(bp, port, 40);
		mux_modem_no_activity_timer(bp, port, 0);
		mux_transmittion_mode(bp, port, FULL_DUPLEX_HARDWIRED);
		mux_hardware_xon_xoff(bp, port, XON_DISABLE, XOFF_DISABLE);
		mux_host_xon_xoff(bp, port, XON_DISABLE, XOFF_DISABLE);
		mux_set_count_always(bp, port);
		ms->counts[port] = TBUFSIZE;
	}		
	mux_end_block(bp);
	ms->cmdq->count = bp - ms->outbuf;

	if (ciosync(ms->cr)) {
		printf("failed to set port characteristics on mux%d\n",
		       unit);
		return EIO;
	}

	/*
	 * finally install the interrupt service routine
	 */
	cc->int_level = SPL_IBIT_CIOMUX;
	(void)procitab(SPLCIOMUX, muxintr, 0, 0);

	/*
	 * signal that this mux is alive and configured
	 */
	ms->alive = 1;

	muxpollon();
	mux_polled = 1;
}

muxscan(unit) 
{
	return 0;
};

void 
muxdisplay(unit)
{
}

muxlayer(unit, layer, devname)
	int *layer;
	char *devname;
{
	if (*layer >= 6)
		return ENXIO;

	*devname++ = 't';
	*devname++ = 't';
	*devname++ = 'y';
	*devname++ = (unit * 6 + *layer) / 10 + '0';
	*devname++ = (unit * 6 + *layer) % 10 + '0';
	*devname++ = '\0';

	return 0;
}

void
muxintr()
{
	/*
	 * we don't do anything for an instrrupt, just service it
	 */
	int unit;

	for (unit = 0; unit < NMUX; unit++) {
		struct cio_chan *cc = mux_softc[unit].cc;
		if (test_cio_are(cc))
			clear_cio_are(cc);
	}
}

muxreset()
{
	/* XXX - do something here */
}

muxopen(dev, flag, ior)
	dev_t dev;
	int flag;
	io_req_t ior;
{
	int board = BOARD(dev);
	int port  = PORT(dev);
	struct mux_softc *ms = &mux_softc[board];
	register struct tty *tp = &ms->tty[port];

	if (board > NMUX || port > 6 || !ms->alive)
		return ENXIO;
		
	tp->t_oproc = muxstart;
	tp->t_stop = muxstop;
	if ((tp->t_state & TS_ISOPEN) == 0) {
		ttychars(tp);
		tp->t_state |= TS_HUPCLS;
		tp->t_ispeed = B9600;
		tp->t_ospeed = B9600;
		tp->t_flags = (ANYP | ECHO);
	}
	
	/*
	 * turn on the receiver and assert DTR and RTS
	 */
	muxctl(ms, port, MUX_ENABLE, 1);
	muxmodem(ms, port, TM_DTR | TM_RTS, 1);

	if (ms->soft_carrier & 1 << port)
		tp->t_state |= TS_CARR_ON;

	return (char_open(dev, tp, flag, ior));
}

muxclose(dev, flag)
	dev_t dev;
{
	int board = BOARD(dev);
	int port  = PORT(dev);
	struct mux_softc *ms = &mux_softc[board];
	register struct tty *tp = &ms->tty[port];
	int s;

	s = spltty();
	simple_lock(&tp->t_lock);

	if (tp->t_state & (TS_HUPCLS | TS_WOPEN) || 
	    (tp->t_state & TS_ISOPEN) == 0) {
		muxmodem(ms, port, TM_DTR | TM_RTS, 1);
		muxctl(ms, port, MUX_ENABLE, 0);
	}

	ttyclose(tp);
	simple_unlock(&tp->t_lock);
	splx(s);
}

muxread(dev, uio)
	dev_t dev;
	struct uio *uio;
{
	struct tty *tp = &mux_softc[BOARD(dev)].tty[PORT(dev)];
	return ((*linesw[tp->t_line].l_read)(tp, uio));
}

muxwrite(dev, uio)
	dev_t dev;
	struct uio *uio;
{
	struct tty *tp = &mux_softc[BOARD(dev)].tty[PORT(dev)];
	return ((*linesw[tp->t_line].l_write)(tp, uio));
}

int
muxportdeath(dev, port)
	dev_t	dev;
	mach_port_t port;
{
	struct tty *tp = &mux_softc[BOARD(dev)].tty[PORT(dev)];
	return (tty_portdeath(tp, port));
}

io_return_t
muxgetstat(dev, flavor, data, count)
	dev_t		dev;
	int		flavor;
	int *		data;		/* pointer to OUT array */
	unsigned int	*count;		/* out */
{
	int board = BOARD(dev);
	int port  = PORT(dev);
	struct mux_softc *ms = &mux_softc[board];
	register struct tty *tp = &ms->tty[port];

	switch (flavor) {
	case TTY_MODEM:
		*data = ms->modem[port];
		*count = 1;
		break;
	default:
		return (tty_get_status(tp, flavor, data, count));
	}
	return (D_SUCCESS);
}

io_return_t
muxsetstat(dev, flavor, data, count)
	dev_t		dev;
	int		flavor;
	int *		data;
	unsigned int	count;
{
	int board = BOARD(dev);
	int port  = PORT(dev);
	struct mux_softc *ms = &mux_softc[board];
	register struct tty *tp = &ms->tty[port];
	int s;

	switch (flavor) {
	case TTY_MODEM:
		if (count < TTY_MODEM_COUNT)
			return (D_INVALID_OPERATION);
		muxmodem(ms, port, *data);
		break;

	case TTY_SET_BREAK:
		muxctl(ms, port, MUX_BREAK, 1);
		break;

	case TTY_CLEAR_BREAK:
		muxctl(ms, port, MUX_BREAK, 0);
		break;

	case TTY_STATUS:
	{
		register int error = tty_set_status(tp, flavor, data, count);
		if (error == 0)
			muxparam(ms, port);
		return (error);
	}

	default:
		return (tty_set_status(tp, flavor, data, count));
	}
	return (D_SUCCESS);
}

muxstart(tp)
	register struct tty *tp;
{
}

muxstop(tp, flag)
	register struct tty *tp;
{
	register int s;

	s = spltty();
	if (tp->t_state & TS_BUSY) {
		if ((tp->t_state&TS_TTSTOP) == 0)
			tp->t_state |= TS_FLUSH;
	}
	splx(s);
}

mux_polled_read(ms)
	register struct mux_softc *ms;
{
	register u_char *bp = ms->inbuf;
#ifdef KGDB
	int need_kgdb = 0;
#endif
	ms->cr->cmdq = ms->read;
	if (ciosync(ms->cr))
		return;

	if (ms->read->count <= ms->stat->count)
		return;

	while (*bp != MUX_ENDBLOCK) {
		register u_int port   = (u_int) *bp++;
		register u_int status = (u_int) *bp++;
		register u_int count  = (u_int) *bp++;
		register struct tty *tp = &ms->tty[port & 0x7f];
			
		/*
		 * check for status packets first
		 */
		if (port & 0x80) {
			register u_int temp;
			register u_int modem;
			int car;
			port &= 0x7f;
		
			switch (status) {
			case MUX_PORT_COUNT:
				ms->counts[port] = TBUFSIZE - (u_int) *bp;
				break;
			
			case MUX_MODEM_STATE:
				temp = (u_int) *bp;
				modem = ms->modem[port] & ~TM_LE;

				if (temp & MUX_CTS)
					modem |= TM_CTS;

				if (temp & MUX_RI)
					modem |= TM_RNG;

				if (temp & MUX_DCD)
					modem |= TM_CAR;

				if (temp & MUX_DSR)
					modem |= TM_DSR;

				if (temp & MUX_RTS)
					modem |= TM_RTS;

				if (temp & MUX_DTR)
					modem |= TM_DTR;

				ms->modem[port] = modem;

				if (ms->soft_carrier & 1 << port)
					car = 1;
				else {
					car = ms->modem[port] & TM_CAR;
				}
				if (car) {
					if ((tp->t_state & TS_CARR_ON) == 0)
						(*linesw[tp->t_line].l_modem)
							(tp, 1);
				} else
					if ((tp->t_state & TS_CARR_ON) == 1)
						(*linesw[tp->t_line].l_modem)
							(tp, 0);
				break;
			}
			bp += count;

		} else {

			if (status & MUX_STATUS_BREAK) {
				(*linesw[tp->t_line].l_rint)(tp->t_breakc,tp);
				bp += count;
				continue;
			}

#ifdef KGDB
			if (*bp == '!' && port == PORT(kgdb_dev) &&
			    ms == &mux_softc[BOARD(kgdb_dev)]) {
				bp++;
				count--;
				need_kgdb = 1;
			}
#endif

			if (tp->t_state & TS_ISOPEN) {
				register int i;
				for (i = 0; i < count; i++)
					(*linesw[tp->t_line].l_rint)(*bp++,tp);
			} else 
				bp += count;
		}
	}

#ifdef KGDB
	if (need_kgdb) {
		extern int kgdb_dummy;

		/*
		 * Awaken the kgdb dummy thread,
		 * which takes a breakpoint in a non-interrupt context.
		 */
		thread_wakeup(&kgdb_dummy);
	}
#endif
}

mux_polled_write(ms)
	register struct mux_softc *ms;
{
	int port;
	int len;
	int actual;
	register u_char *bp;
	u_char *sbp;

	bp = ms->bp;

	for (port = 0; port < 6; port++) {
		register struct tty *tp = &ms->tty[port];
		if ((tp->t_state & TS_ISOPEN) == 0)
			continue;

		tp->t_state &=~ TS_FLUSH;

		len = tp->t_outq.c_cc > ms->counts[port] ?
			ms->counts[port] : tp->t_outq.c_cc;
		if (len == 0)
			continue;
		sbp = bp;
		mux_start_data_block(bp, port, len);

		actual = q_to_b(&tp->t_outq, bp, len);
		if (actual != len) {
			mux_start_data_block(sbp, port, actual);
			len = actual;
		}
		bp += len;
		ms->counts[port] -= len;
	}

	if (bp == ms->outbuf)
		return;

	mux_end_block(bp);

	ms->write->count = bp - ms->outbuf;
	ms->cr->cmdq = ms->write;
	ciosync(ms->cr);
	ms->bp = ms->outbuf;
}

mux_poll()
{
	register int mux;
	for (mux = 0; mux < NMUX; mux++) {
		register struct mux_softc *ms = &mux_softc[mux];
		if (!ms->alive)
			continue;

		mux_polled_read(ms);
		mux_polled_write(ms);
	}
	if (mux_polled)
		muxpollon();
}

muxmodem(ms, port, data, set)
	struct mux_softc *ms;
{
	int s;
	register u_char *bp;

	s = spltty();
	bp = ms->bp;

	if (set) {
		if (data & TM_DTR)
			mux_set_dtr_line(bp, port);

		if (data & TM_RTS)
			mux_set_rts_line(bp, port);
	} else {
		if (data & TM_DTR)
			mux_clear_dtr_line(bp, port);

		if (data & TM_RTS)
			mux_clear_rts_line(bp, port);
	}
	ms->bp = bp;
	splx(s);
}

muxctl(ms, port, data, set)
	struct mux_softc *ms;
{
	int s;
	register u_char *bp;

	s = spltty();
	bp = ms->bp;

	if (set) {
		if (data & MUX_ENABLE) {
			mux_enable_receiver(bp, port);
			ms->modem[port] |= TM_LE;
		}

		if (data & MUX_BREAK)
			mux_start_break(bp, port);
	} else {
		if (data & MUX_ENABLE) {
			mux_disable_receiver(bp, port);
			ms->modem[port] &= ~TM_LE;
		}

		if (data & MUX_BREAK)
			mux_stop_break(bp, port);
	}
	ms->bp = bp;
	splx(s);
}

u_int baudrates[] = {
	BAUD_9600,			/* B0 */
	BAUD_50,			/* B50 */
	BAUD_75,			/* B75 */
	BAUD_110,			/* B110 */
	BAUD_134,			/* B134 */
	BAUD_150,			/* B150 */
	BAUD_300,			/* B200 */
	BAUD_300,			/* B300 */
	BAUD_600,			/* B600 */
	BAUD_1200,			/* B1200 */
	BAUD_1800,			/* B1800 */
	BAUD_2400,			/* B2400 */
	BAUD_4800,			/* B4800 */
	BAUD_9600,			/* B9600 */
	BAUD_9600,			/* EXTA */
	BAUD_9600			/* EXTB */
};

muxparam(ms, port)
	struct mux_softc *ms;
{
	int s;
	register u_char *bp;
	register struct tty *tp = &ms->tty[port];
	int mask = 0;

	s = spltty();
	bp = ms->bp;

	if (tp->t_ispeed == 0) {
		mux_disable_receiver(bp, port);
		ms->modem[port] &= ~TM_LE;
	} else 
		mux_baudrate(bp, port, baudrates[tp->t_ispeed]);

#ifndef	TIP_DEBUGGING
	switch(tp->t_flags & TF_ANYP) {
	case TF_ANYP:
		mask |= PARITY_NONE;
		break;

	case TF_EVENP:
		mask |= PARITY_EVEN;
		break;

	case TF_ODDP:
		mask |= PARITY_ODD;
		break;
	}

	if (tp->t_flags & TF_LITOUT)
		mask |= BITS_8;
	else
		mask |= BITS_7;

	if (tp->t_ispeed == B110)
		mask |= STOP_2;
	else
		mask |= STOP_1;

	mux_char_length(bp, port, mask);

#endif	TIP_DEBUGGING

	ms->bp = bp;
	splx(s);
}

/*
 * The following routines allow the mux to act as the console (KGDB uses the
 * console getc/putc interface as well)
 *
 * The console device will always be mux0 port 0
 */
#include <hpdev/cons.h>

muxcnprobe(cp)
	struct consdev *cp;
{
	int unit;
	
	if (!mux_softc[0].alive) {
		cp->cn_pri = CN_DEAD;
		return;
	}
	unit = 0;				/* XXX */
	cp->cn_pri = CN_NORMAL;
	cp->cn_dev = unit;
	cp->cn_tp  = &(mux_softc[BOARD(unit)].tty[PORT(unit)]);
}

muxcninit(cp)
	register struct consdev *cp;
{
	/*
	 * not much to do initializing the console. We'll defer all real 
	 * intialization until someone tries to use the port. This way KGDB 
	 * won't have to do anything special to initialize its port.
	 */
	return 0;
}

muxcnopen(dev)
	dev_t dev;
{
	int board = BOARD(dev);
	int port  = PORT(dev);
	struct mux_softc *ms = &mux_softc[board];
	register struct tty *tp = &ms->tty[port];

	/*
	 * set the soft carrier if we are using this for getc and putc
	 */
	ms->soft_carrier |= 1 << port;

	/*
	 * call the normal open routine to initialize character buffers
	 * since we have a soft carrier we don't actually need an io_req 
	 * structure
	 */
	muxopen(dev, 0, 0);

	/*
	 * don't hang up on close
	 */
	tp->t_state &= ~TS_HUPCLS;
	
	/*
	 * We only want to do this for the console port. 
	 */
	if (PORT(dev) == 0)
		tp->t_flags |= XTABS|CRMOD;
}

muxcngetc(dev)
	dev_t dev;
{
	int board = BOARD(dev);
	int port  = PORT(dev);
	struct mux_softc *ms = &mux_softc[board];
	register struct tty *tp = &ms->tty[port];
	int s;
	int c;

	if ((tp->t_state & TS_ISOPEN) == 0)
		muxcnopen(dev);

	/*
	 * OK, we're just going to steal the characters right off the tty
	 * input queue. If there are no characters then we'll loop until
	 * we get one. We'll also use splhigh so that we can get characters
	 * for KGDB in interrupt service routines without the possibility
	 * of lowering spl in an illegal way.
	 */
	if (mux_polled) {
		mux_polled = 0;
		muxpolloff();
	}

	while(1) {
		s = splhigh();
		if (tp->t_inq.c_cc > 0)
			break;
		splx(s);
		mux_poll();
	}
		
	c = getc(&tp->t_inq);
	splx(s);

	mux_polled = 1;
	muxpollon();

	return c;
}

muxcnputc(dev, c)
	dev_t dev;
{
	int board = BOARD(dev);
	int port  = PORT(dev);
	struct mux_softc *ms = &mux_softc[board];
	register struct tty *tp = &ms->tty[port];
	int s;

	if ((tp->t_state & TS_ISOPEN) == 0)
		muxcnopen(dev);

	if (mux_polled) {
		mux_polled = 0;
		muxpolloff();
	}

	c &= 0xff;

	while (1) {
		s = splhigh();
		simple_lock(&tp->t_lock);

		if (putc(c, &tp->t_outq) == 0)
			break;

		simple_unlock(&tp->t_lock);
		splx(s);

		/*
		 * can't fit the character on the output queue, try writing
		 * some more characters out
		 */
		mux_poll();
	}

	/*
	 * write out the character we have
	 */
	mux_poll();

	simple_unlock(&tp->t_lock);
	splx(s);
	mux_polled = 1;
	muxpollon();
}
#endif	NMUX > 0
