/* 
 * Copyright (c) 1992, 1994 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 * All rights reserved.
 *
 * Permission to use, copy, modify and distribute this software is hereby
 * granted provided that (1) source code retains these copyright, permission,
 * and disclaimer notices, and (2) redistributions including binaries
 * reproduce the notices in supporting documentation, and (3) all advertising
 * materials mentioning features or use of this software display the following
 * acknowledgement: ``This product includes software developed by the
 * Computer Systems Laboratory at the University of Utah.''
 *
 * THE UNIVERSITY OF UTAH AND CSL ALLOW FREE USE OF THIS SOFTWARE IN ITS "AS
 * IS" CONDITION.  THE UNIVERSITY OF UTAH AND CSL DISCLAIM ANY LIABILITY OF
 * ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING FROM THE USE OF THIS SOFTWARE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: db_hp800_symtab.c 1.4 94/12/14$
 */
/*
 * This file written by Patrick Dalton, 6/4/92
 */

#include <mach/mach_interface.h>
#include <mach/mach_user_internal.h>
#include <mach/mach_port_internal.h>

#include <boot_ufs/file_io.h>
#include <boot_ufs/loader_info.h>

#include <ddb/stab.h>
#include <ddb/nlist.h>

#include <machine/db_trace.h>
#include <machine/db_symtab.h>

extern char end[];
char       *esym = end;

/*
 * get_symtab:  
 *
 * This function reads the header of a som binary, then
 * uses it to locate the beginning and size of the symbol
 * table.  It returns FALSE if the magic sym is not correct
 * (i.e. believes this is not a som binary) of if there is 
 * an io problem.
 */
boolean_t
get_symtab(fp, symoff_p, symsize_p, header, header_size)
     struct file     *fp;
     vm_offset_t     *symoff_p;
     vm_size_t       *symsize_p;
     char            header[];
     vm_size_t       *header_size;
{
  SCNHDR  symtab_info[2];
  unsigned int null_info;

  kern_return_t result;
  vm_size_t     resid;

  boolean_t read_symtab_info_records();

  if (!read_symtab_info_records(fp, symtab_info, &null_info))
    return FALSE;
  else if (null_info)
    {
      *symsize_p   = 0;
      *header_size = 0;
      return TRUE;
    }
  else
    {
      *symoff_p  = symtab_info[0].file_loc_init_value;
      *symsize_p = symtab_info[0].initialization_length + symtab_info[1].initialization_length;

      *(unsigned int *)header = symtab_info[1].initialization_length;
      *header_size            = sizeof(unsigned int);

      return TRUE;
    }
}


/*
 * read_symtab:
 * Read in the symbol table information putting symbol information first
 * and the string information second.
 */
boolean_t
read_symtab(fp, symtab, sym_off, sym_size, sym_hdr_size)
     struct file *fp;
     char        *symtab;
     vm_offset_t sym_off;
     vm_size_t   sym_size;
     vm_size_t   sym_hdr_size;
{
  SCNHDR symtab_info[2];  /* info 0 points to the strings; info 1 points to the symbols */
  unsigned int null_info;

  kern_return_t result;
  vm_size_t     resid;

  boolean_t read_symtab_info_records();

  if (!read_symtab_info_records(fp, symtab_info, &null_info))
    return FALSE;
  else if (null_info)  /* no debug information to be found */
    {
      /* read_symtab should not be called in this case */
      return TRUE;
    }
  else
    {
      result = 
	read_file(fp, symtab_info[1].file_loc_init_value, symtab + sym_hdr_size, 
		  symtab_info[1].initialization_length, &resid);
      if (result || resid)
	return FALSE;
      
      result =
	read_file(fp, symtab_info[0].file_loc_init_value, 
		  symtab + sym_hdr_size + symtab_info[1].initialization_length , 
		  symtab_info[0].initialization_length, &resid);
      if (result || resid)
	return FALSE;

      return TRUE;
    }
}


/*
 * read_symtab_info_records:
 * This function searches through the som file for the subspace records
 * referring to the GDB debug information.
 */
boolean_t
read_symtab_info_records(fp, subspaces, null_spaces)
     struct file *fp;
     SCNHDR subspaces[2];
     unsigned int *null_spaces;
{
  FILHDR  filhdr;
  SPAHDR  space;
  char   *space_name;
  vm_offset_t subspace_entry_offset;

  kern_return_t result;
  vm_size_t     resid;
  unsigned int i;

  /* read file header */
  result = 
    read_file(fp, 0, (char *)&filhdr, FILHSZ, &resid);
  if (result || resid)
    return FALSE;

  if (filhdr.system_id != HP9000S800_ID)
    return FALSE;

  /* allocate space for space names */
  result =
    vm_allocate(mach_task_self(), &space_name, filhdr.space_strings_size, TRUE);
  if (result != KERN_SUCCESS)
    {
      db_printf("[ Unable to allocate space for space name strings]\n");
      return FALSE;
    }
  
  /* read space names */
  result = 
    read_file(fp, filhdr.space_strings_location, space_name,
	      filhdr.space_strings_size, &resid);
  if (result || resid)
    return FALSE;

  /* find debugging symbols space */
  for(i=0; i<filhdr.space_total; i++)
    {
      result =
	read_file(fp, filhdr.space_location+i*SPAHSZ, 
		  (char *)&space, SPAHSZ, &resid);
      if (result || resid)
	return FALSE;
      if (strcmp(DEBUG_SYMBOL_SPACE_NAME, space_name + space.STR_INDEX) == 0)
	break;
    }
  (void)vm_deallocate(mach_task_self(), &space_name, filhdr.space_strings_size);

  if (i == filhdr.space_total) /* there is no debug symbol information */
    {
      *null_spaces = 1;
      return TRUE;
    }

  if (space.subspace_index > filhdr.subspace_total)
    {
      db_printf( "[ Bad subspace index in space %s ]\n", 
		DEBUG_SYMBOL_SPACE_NAME);
      return FALSE;
    }

  if (space.subspace_index < 0)
    {
      *null_spaces = 1;
      return TRUE;
    }

  /* read subspace table describing the location of debugging symbols */
  subspace_entry_offset = 
    space.subspace_index * SCNHSZ + filhdr.subspace_location;

  result = 
    read_file(fp, subspace_entry_offset, (char *)subspaces, 2*SCNHSZ, &resid);
  if (result || resid)
    return FALSE;

  *null_spaces = 0;
 
  return TRUE;
}



#define IS_LOCAL_SYM(sym) \
  (((sym)->n_type == N_LSYM) || ((sym)->n_type == N_RSYM))

#define IS_REG_SYM(sym) \
  (((sym)->n_type == N_RSYM) ? (sym)->n_value : 0)

#define IS_FUNCTION(sym) \
  ((sym)->n_type == N_FUN)

#define IS_OPEN_LEVEL_SYM(sym) \
  ((sym)->n_type == N_LBRAC)

#define IS_CLOSE_LEVEL_SYM(sym) \
  ((sym)->n_type == N_RBRAC)

#define IS_LINENUM(sym) \
  ((sym)->n_type == N_SLINE)


struct nlist *find_first_local_sym(struct nlist *start_sym, 
				   unsigned int *nest_level);
boolean_t is_param(struct nlist *sym);
boolean_t is_local_param(struct nlist *sym);


/*
 * get_next_arg_sym:
 *
 * This function looks through an nlist symbol table for the
 * next parameter symbol.
 */
boolean_t
get_next_arg_sym(start_sym, nest_level, arg_sym)
     db_sym_t start_sym, *arg_sym;
     unsigned int *nest_level;
{
  db_symtab_t *symtab        = db_last_symtab;
  struct nlist *end_of_syms  = (struct nlist *)(symtab->end);
  struct nlist *sym          = (struct nlist *)start_sym;

  if (IS_FUNCTION(sym))
    sym++;

  *arg_sym = (db_sym_t)sym;
  return is_param(sym);
}


/*
 * get_next_local_sym:
 *
 * This function looks through an nlist symbol table for the
 * next local variable symbol.  If it starts out in the parameters
 * it steps to the end of the local version of the parameters.
 */
db_sym_t
get_next_local_sym(start_sym, is_reg, nest_level)
     db_sym_t start_sym;
     unsigned int *is_reg, *nest_level;
{
  db_symtab_t *symtab       = db_last_symtab;
  struct nlist *end_of_syms = (struct nlist *)(symtab->end);
  struct nlist *sym         = (struct nlist *)start_sym;

  if (!IS_LOCAL_SYM(sym) || is_local_param(sym))
    sym = find_first_local_sym(sym, nest_level);

  for (; sym != end_of_syms; sym++)
    {
      if(IS_LOCAL_SYM(sym))
	{
	  if (is_reg)
	    *is_reg = IS_REG_SYM(sym);
	  return (db_sym_t)sym;
	}
      else if (IS_OPEN_LEVEL_SYM(sym))
	*nest_level ++;
      else if (IS_CLOSE_LEVEL_SYM(sym))
	*nest_level --;
    }

  return DB_SYM_NULL;
}
	    

/*
 * find_first_local_sym:
 *
 * This function scans down an nlist for the first varible
 * symbol that is not also a parameter.
 */
struct nlist *
find_first_local_sym(struct nlist *start_sym, unsigned int *nest_level)
{
  db_symtab_t *symtab       = db_last_symtab;
  struct nlist *end_of_syms = (struct nlist *)(symtab->end);
  struct nlist *sym         = (struct nlist *)start_sym;

  for (;sym != end_of_syms; sym++)
    {
      if(IS_LOCAL_SYM(sym) && is_local_param(sym))
	return sym;
      else if (IS_OPEN_LEVEL_SYM(sym))
	*nest_level ++;
      else if (IS_CLOSE_LEVEL_SYM(sym))
	*nest_level --;
    }
  
  return sym;
}


/*
 * is_local_param:
 *
 * This function finds out if sym is really a parameter
 * to the function.
 */
boolean_t
is_local_param(struct nlist *sym)
{
  unsigned int len;
  register struct nlist *cur_sym = sym;
  register char *sym_name = sym->n_un.n_name;
  char *index();

  while(IS_LOCAL_SYM(cur_sym)) cur_sym--;

  len = (unsigned int)(sym->n_un.n_name - index(sym->n_un.n_name, ':'));

  for(; !IS_FUNCTION(cur_sym); cur_sym--)
    if (strncmp(cur_sym->n_un.n_name, sym_name, len))
      return TRUE;

  return FALSE;
}


/*
 * is_param:
 * 
 * A function that determines whether a symbol is a function
 * parameter or not.
 */
boolean_t
is_param(struct nlist *sym)
{
  char *param_indicator;
  char *index();

  if (sym->n_type == N_PSYM)
    return TRUE;

  if (sym->n_type == N_LSYM)
    {
      param_indicator = index(sym->n_un.n_name, ':');

      if(*++param_indicator == 'p')
	return TRUE;
    }

  return FALSE;
}


/*
 * get_num_func_args:
 * 
 * a function that counts the number of parameters a function should recieve.
 */
get_num_func_args(start_sym)
     db_sym_t start_sym;
{
  struct nlist *sym = (struct nlist *)start_sym;
  int count = 0;
  
  if (!IS_FUNCTION(sym))
    return 4;

  for(sym++; is_param(sym); sym++, count++);

  return count;
}
 

/*
 * get_num_func_locals:
 *
 * A function that counts the number of locals a funtion has
 */
get_num_func_locals(start_sym, num_params, max_num_locals)
     db_sym_t start_sym;
     int num_params, max_num_locals;
{
  struct nlist *sym = (struct nlist *)start_sym;
  int count, i;

  if (!IS_FUNCTION(sym))
    return max_num_locals;
  
  for(i=2*num_params; i>0; i++, sym++);

  while (!IS_LINENUM(sym))
    if (IS_LOCAL_SYM(sym))
      count ++;

  return count;
}


/*
 * index:
 *
 * This is a local version of the string function index.
 */
static char *
index(str, ch)
     char *str, ch;
{
  for(; *str; str++)
    if (*str == ch)
      return str;
}


/*
 * strncmp:
 *
 * This is a local verson of the string function strncmp.
 */
static
strncmp(s1, s2, count) 
     char *s1, *s2;
     int count;
{
  while (*s1 && *s2 && *s1++ == *s2++ && count--);

  if (!count || !*s1 && !*s2) return 0;
  if (!*s1)   return 1;
  if (!*s2)   return -1;
  if (*s1 > *s2 )
    return 1;
  else
    return -1;
}
      

  



     
      
