/*
 * "Stub" to allow remote cpu to debug over a serial line using gdb.
 *
 * Copyright (c) 1990 Regents of the University of California.
 * Copyright (c) 1990,1991,1992,1994 The University of Utah and
 * the Computer Systems Laboratory (CSL).
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Berkeley and the Center
 * for Software Science at the University of Utah.  The names of the
 * Universities may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * CSL requests users of this software to return to csl-dist@cs.utah.edu any
 * improvements that they make and grant CSL redistribution rights.
 *
 * 	Utah $Hdr: kgdb_stub.c 1.21 94/11/10$
 */

/*
 * The following gdb commands are supported:
 *
 * command          function                               Return value
 *
 *    g             return the value of the CPU registers  hex data or ENN
 *    G             set the value of the CPU registers     OK or ENN
 *
 *    mAA..AA,LLLL  Read LLLL bytes at address AA..AA      hex data or ENN
 *    MAA..AA,LLLL: Write LLLL bytes at address AA.AA      OK or ENN
 *
 *    c             Resume at current address              SNN   ( signal NN)
 *    cAA..AA       Continue at address AA..AA             SNN
 *
 *    s             Step one instruction                   SNN
 *    sAA..AA       Step one instruction from AA..AA       SNN
 *
 *    k             kill
 *
 *    ?             What was the last sigval ?             SNN   (signal NN)
 *
 * All commands and responses are sent with a packet which includes a
 * checksum.  A packet consists of
 *	 $<packet info>#<checksum>.
 * where
 * <packet info> :: <characters representing the command or response>
 * <checksum>    :: < two hex digits computed as modulo 256 sum of <packetinfo>>
 *
 * When a packet is received, it is first acknowledged with either '+' or '-'.
 * '+' indicates a successful transfer.  '-' indicates a failed transfer.
 */

/*
 * Addendum on the Mach 3.0 server debugging stub:
 *
 * For some reason Mach 3 puts kernel functionality into user mode servers.
 * Unfortunately local debugging of user servers is a big problem, since a
 * huge amount of code in (e.g.) the Mach single server must be working in
 * order to support a debugger, and even then you don't dare debug a
 * server that the debugger is currently using.
 *
 * Therefore we have extended the stub to support multi-thread, multi-task
 * remote debugging.  The interface is simple and somewhat naive.  The
 * basic unit of debugging is the task; for each task being debugged, the
 * user needs a gdb session.  The first task to be debugged is always the
 * kernel.  From the kernel session, it is possible to switch explicitly
 * to new sessions.  Once a session has started, the stub automatically
 * switches to it when a thread in the given task requests service, or
 * when the user's gdb makes a request.
 *
 * On the local end, a multiplexer program mediates access to the remote
 * stub.  Gdb sessions attach to local pseudo-ttys; the multiplexer hides
 * all the overhead of switching in the stub, so gdb itself needs no
 * changes.  Only the multiplexer opens the hard line to the remote
 * system.  The multiplexer uses an 'X' packet, new to the protocol, to
 * request a switch to a different session.  The kernel sends the
 * multiplexer an X packet when the remote end requests a switch.
 *
 * Explicit task switching is accomplished by setting the kernel variable
 * kgdb_context in the kernel gdb session.  This is normally used once for
 * each new gdb session to get started.  The value of kgdb_context can be
 * one of three things: a thread pointer, a task pointer or 0.  If the
 * user provides a task pointer, the stub automatically selects a thread
 * in the task.  Threads are selected merely for their register state;
 * actions such as installing breakpoints are not limited to particular
 * threads.  An important note: an explicit task switch does NOT suspend
 * the selected thread or task, so it's best to pick a thread that is
 * already blocked.  Also, the selected thread is generally NOT the
 * current thread in the kernel gdb session.
 *
 * Kernel debugging still works much the same way that it used to, and it
 * is possible to perform kernel debugging without starting up the
 * multiplexer.  The stub takes care not to send X packets until it sees
 * the first request to switch sessions, so that it does not confuse naive
 * gdb's.  Kernel breakpoints that are executed by user threads are
 * handled by the kernel gdb session, as a special case.  It is possible
 * to run gdb commands from the interrupt stack without real threads being
 * available, although in the usual case the stub works with thread
 * operations internally and thus debugging context switch code can
 * confuse the debugger (and possibly cause hangs or panics).  Debugging
 * on the interrupt stack works but limits your options -- you may not
 * have any other gdb session in progress, and you may not take a
 * breakpoint in another interrupt routine or in a user thread while
 * executing on the interrupt stack.
 *
 * The internals of the stub have changed quite radically (and the stub
 * has tripled in size, unfortunately).  There is now a separate kgdb
 * thread that takes requests for service from other threads.  When a
 * thread calls kgdb_trap(), it normally blocks and requests service from
 * the kgdb thread.  The kgdb thread operates on the blocked thread and
 * eventually releases it, with a continue or single step operation.
 * Threads may block if the kgdb thread is already busy; in this
 * situation, the kgdb thread will pick up the next blocked thread when it
 * releases the active thread.  The kgdb thread may switch away from a
 * thread without releasing it; in this case the thread goes back into the
 * pool of waiting threads and takes its turn like the others.
 *
 * An exception to this scheme occurs when an interrupt routine needs kgdb
 * service (for example, we hit a breakpoint in a driver).  In this case,
 * the thread that took the interrupt handles the kgdb requests directly,
 * and switching is not permitted.  One common case is not handled on the
 * interrupt stack -- if the serial port interrupt routine detects the
 * kgdb exception character ('!'), it wakes up a dummy thread that takes a
 * breakpoint in its own context.  This means that all the regular
 * multi-threading features are available after interrupting a kernel with
 * a ^C from the kernel gdb session.
 *
 * Since it is now possible to read data from any selected address space,
 * we augmented the old kernacc() code to deal with multiple VM maps.
 * Kernel accesses are still done the old way, but we handle user space
 * accesses by mapping the user page(s) into the kernel.  Currently we
 * access the user page directly for both reading and writing; this means
 * that we don't cause COW faults on COW pages, and we propagate altered
 * data to all tasks that share the page.  (Don't set breakpoints in
 * shared text...)  As a small optimization, we check new accesses to see
 * if they can be satisfied by the previous mapping and preserve the
 * mapping if we succeed.
 *
 * This code is unsupported, unmaintained, unloved, etc.  Just thought I'd
 * tell you now...
 *
 * Donn Seeley, University of Utah CSL, 1/17/91
 */

#ifndef lint
static char rcsid[] = "$Header: /n/fast/usr/lsrc/mach/CVS/mach4-parisc/kernel_unused/parisc/kgdb_stub.c,v 1.22 1994/12/14 19:38:34 law Exp $";
#endif

#ifdef KGDB

#include <kern/thread.h>
#include <kern/processor.h>
#include <mach/thread_status.h>
#include <mach/vm_param.h>
#include <machine/trap.h>
#include <machine/psl.h>
#include <machine/spl.h>
#include <sys/reboot.h>

extern void printf();
extern void bcopy();
extern int kgdbacc();
extern void chgkprot();

extern int istackptr;

/*
 * There are 520 bytes of registers on a PA 1.1 machine
 */
#define NUMREGBYTES 520
/*
 * BUFMAX defines the maximum number of characters in inbound/outbound
 * buffers.  At least NUMREGBYTES*2 are needed for register packets.
 */
#define BUFMAX 2048

/*
 * Bits for the 'kgdb_trace' field in the pcb.
 */
#define	KGDB_TRACE	1
#define	KGDB_STEP	2
#define	KGDB_INPROGRESS	4

#ifndef KGDBDEV
#define KGDBDEV -1
#endif
#ifndef KGDBRATE
#define KGDBRATE 9600
#endif

int kgdb_dev = KGDBDEV;		/* remote debugging device (-1 if none) */
int kgdb_rate = KGDBRATE;	/* remote debugging baud rate */
int kgdb_debug_init = 1;	/* != 0 waits for remote at system init */
int kgdb_debug = 0;		/* > 0 prints command & checksum errors */
				/* > 1 prints packet traffic */
thread_t kgdb_context = (thread_t) -1;	/* kgdb's idea of the current thread */
int kgdb_lock = 0;		/* set if already executing in kgdb_trap() */
struct hp800_saved_state *kgdb_ssp;	/* kgdb_thread() parameter */
int kgdb_type;			/* kgdb_thread() parameter */
int kgdb_spl;			/* kgdb_thread() parameter */
thread_t kgdb_command();	/* kgdb_thread() parameter */
thread_t kgdb_pick_thread();

/*
 * Some selected Unix signals, for gdb's sake.
 */
#define	SIGINT		2
#define	SIGTRAP		5
#define	SIGSEGV		11
#define	SIGALRM		14

#include <sys/types.h>	/* XXX what the hell?!? */
#include <hpdev/cons.h>

extern struct consdev constab[];

#define GETC	\
	(constab[major(kgdb_dev)].cn_getc ? \
		(*constab[major(kgdb_dev)].cn_getc)(kgdb_dev, TRUE) : 0)
#define PUTC(c)	{ \
	if (constab[major(kgdb_dev)].cn_putc) \
		(*constab[major(kgdb_dev)].cn_putc)(kgdb_dev, c); \
}

static char hexchars[] = "0123456789abcdef";


static char inbuffer[BUFMAX];
static char outbuffer[BUFMAX];

extern void kgdb_break();

/*
 * this routine is called by the mux initialization code as soon as the
 * mux is configured.
 */
kgdb_mux_init()
{
	extern int boothowto;
	int port = minor(kgdb_dev);
	
	if (boothowto & RB_KDB) {
		boothowto &= ~RB_KDB;
		if (kgdb_debug_init) {
			printf("tty%02x: kgdb waiting...", port);
			kgdb_break();
			printf(" connected.\n");
		} else
			printf("tty%02x: kgdb enabled\n", port);
	}
	else
		kgdb_dev = -1;	/* Must do this! LBS */
}


static /* inline */ int 
hex(ch)
	char ch;
{
	if ((ch >= '0') && (ch <= '9'))
		return (ch - '0');
	if ((ch >= 'a') && (ch <= 'f'))
		return (ch - ('a' - 10));
	return (0);
}

/* scan for the sequence $<data>#<checksum>     */
static void 
getpacket(buffer)
	char *buffer;
{
	unsigned char checksum;
	unsigned char xmitcsum;
	int i;
	int count;
	char ch;

	do {
		/*
		 * wait around for the start character, ignore all other
		 * characters
		 */
		while ((ch = GETC) != '$')
			;
		checksum = 0;
		count = 0;
		xmitcsum = 1;

		/* now, read until a # or end of buffer is found */
		while (count < BUFMAX) {
			ch = GETC;
			if (ch == '#')
				break;
			checksum = checksum + ch;
			buffer[count] = ch;
			count = count + 1;
		}
		buffer[count] = 0;

		if (ch == '#') {
			xmitcsum = hex(GETC) << 4;
			xmitcsum += hex(GETC);
			if (kgdb_debug && (checksum != xmitcsum)) {
				printf(
			"bad checksum.  My count = 0x%x, sent=0x%x. buf=%s\n",
					checksum, xmitcsum, buffer);
			}
			if (checksum != xmitcsum) {
				PUTC('-');	/* failed checksum */
			} else {
				PUTC('+');	/* successful transfer */
				/*
				 * if a sequence char is present, reply the
				 * sequence ID
				 */
				if (buffer[2] == ':') {
					PUTC(buffer[0]);
					PUTC(buffer[1]);
					/* remove sequence chars from buffer */
					for (i = 3; i <= count; ++i)
						buffer[i - 3] = buffer[i];
				}
			}
		}
	} while (checksum != xmitcsum);

	if (kgdb_debug > 1)
		printf("getpacket: %s\n", buffer);
}

/*
 * send the packet in buffer.  The host gets one chance to read it.  This
 * routine does not wait for a positive acknowledge.
 */
static void 
putpacket(buffer)
	char *buffer;
{
	unsigned char checksum;
	int count;
	char ch;

	if (kgdb_debug > 1)
		printf("putpacket: %s\n", buffer);

	/* $<packet info>#<checksum>. */
	do {
		PUTC('$');
		checksum = 0;
		count = 0;

		while (ch = buffer[count]) {
			PUTC(ch);
			checksum += ch;
			count += 1;
		}
		PUTC('#');
		PUTC(hexchars[checksum >> 4]);
		PUTC(hexchars[checksum & 15]);

	} while (0);	/* (GETC != '+'); */

}

static /* inline */ void 
debug_error(format, parm)
	char *format, *parm;
{
	if (kgdb_debug)
		printf(format, parm);
}

/*
 * Convert at most 'dig' digits of hex data in buf into a value.
 * Stop on non-hex char.  Return a pointer to next char in buf.
 */
static char *
hex2val(buf, val, dig)
	char *buf;
	int *val;
	int dig;
{
	int i, v;
	char ch;

	v = 0;
	for (i = dig; --i >= 0; ) {
		ch = *buf++;
		if ((ch >= '0') && (ch <= '9'))
			v = (v << 4) | (ch - '0');
		else if ((ch >= 'a') && (ch <= 'f'))
			v = (v << 4) | (ch - ('a' - 10));
		else {
			--buf;
			break;
		}
	}
	*val = v;
	return (buf);
}

/*
 * convert the integer value 'val' into 'dig' hex digits, placing
 * result in buf.  Return a pointer to the last char put in buf (null).
 */
static char *
val2hex(buf, val, dig)
	char *buf;
	int val;
	int dig;
{
	for (dig <<= 2; (dig -= 4) >= 0; )
		*buf++ = hexchars[(val >> dig) & 0xf];
	*buf = 0;
	return (buf);
}

/*
 * convert the memory pointed to by mem into hex, placing result in buf.
 * return a pointer to the last char put in buf (null).
 */
static char *
mem2hex(buf, mem, count)
	char *buf;
	char *mem;
	int count;
{
	if ((count & 1) || ((int)mem & 1)) {
		u_char ch;

		while(--count >= 0) {
			ch = *mem++;
			*buf++ = hexchars[ch >> 4];
			*buf++ = hexchars[ch & 15];
		}
	} else {
		u_short s;
		u_short *mp = (u_short *)mem;

		for (count >>= 1; --count >= 0; ) {
			s = *mp++;
			*buf++ = hexchars[(s >> 12) & 15];
			*buf++ = hexchars[(s >> 8) & 15];
			*buf++ = hexchars[(s >> 4) & 15];
			*buf++ = hexchars[s & 15];
		}
	}
	*buf = 0;
	return (buf);
}

/*
 * Convert the hex array pointed to by buf into binary to be placed in mem.
 * Return a pointer to next char in buf.
 */
static char *
hex2mem(buf, mem, count)
	char *buf;
	char *mem;
	int count;
{
	int i;
	unsigned char ch;

	for (i = 0; i < count; ++i) {
		ch = hex(*buf++) << 4;
		ch = ch + hex(*buf++);
		*mem++ = ch;
	}
	return (buf);
}

/*
 * Translate a trap number into a unix compatible signal value.
 * (gdb only understands unix signal numbers).
 */
static int 
computeSignal(type)
	int type;
{
	int sigval;

	/*
	 * For use internal to kgdb_stub.c, to send specific signals to gdb.
	 */
	if (type < 0)
		return -type;

	/*
	 * We have to imitate a little bit of the Unix trap() code here...
	 */
	switch (type) {
	case I_BRK_INST:	/* single step w/o branch, or breakpoint */
	case I_TAKEN_BR:	/* single step w/ branch */
	case I_DMEM_BREAK:	/* trace writes to a particular page */
	case I_PAGE_REF:	/* trace references to a particular page */
		sigval = SIGTRAP;
		break;

	case I_RECOV_CTR:	/* beats me what we should return for this */
		sigval = SIGALRM;	/* XXX */
		break;

	case I_HPRIV_XFR:
	case I_LPRIV_XFR:	/* ditto */
		sigval = SIGSEGV;
		break;

	default:
		printf("computeSignal: unsupported trap number %d\n", type);
		panic("unsupported trap number for kgdb");
	}
	return (sigval);
}

/*
 * This routine does two things: it converts a Mach saved_state structure
 * into the structure that gdb and Unix expect to see, and at the same time
 * it changes the data into hex and puts it into the provided buffer.
 */
void
send_state(cp, ssp)
	register char *cp;
	register struct hp800_saved_state *ssp;
{
	int x;
	int zero = 0;

#define	M2H(m, l)							\
	do {								\
		cp = mem2hex(cp, (char *) &ssp->m, l * sizeof (int));	\
	} while(0)
#define	SKIPN(n)							\
	do {								\
		int i;							\
		for (i = 0; i < n; i++)					\
			cp = mem2hex(cp, (char *) &zero, sizeof zero);	\
	} while(0)
#define	C2H(c)								\
	do {								\
		x = mfctl(c);						\
		cp = mem2hex(cp, (char *) &x, sizeof x);		\
	} while(0)							\

	/*
	 * we need to send the saved state structure as it appears in BSD
	 */
	M2H(flags, 1);			/* ss_flags */
	M2H(r1,31);			/* ss_gr1 - ss_gr31 */
	M2H(cr11, 1);			/* ss_cr11 */
	M2H(iioq_head, 1);		/* ss_pcog_head */
	M2H(iisq_head, 1);		/* ss_pcsq_head */
	M2H(iioq_tail, 1);		/* ss_pcoq_tail */
	M2H(iisq_tail, 1);		/* ss_pcsq_tail */
	M2H(cr15, 1);			/* ss_cr15 */
	M2H(cr19, 4);			/* ss_cr{19,20,21,22} */
	SKIPN(1);			/* ss_goto */
	M2H(sr4, 1);			/* ss_sr4 */
	M2H(sr0, 4);			/* ss_sr{0,1,2,3} */
	M2H(sr5, 6);			/* ss_sr{5,6,7}, ss_cr{0,8,9} */
	C2H(10);			/* ss_cr10 */
	M2H(cr12, 2);			/* ss_cr{12,13} */
	SKIPN(7);			/* ss_cr{24,25,26}, 
					 * ss_mpsfu_{high, low, ovflo}
					 * ss_pad
					 */
	M2H(fr0, 32);			/* ss_fr0 - ss_fr15 */

#undef	M2H
#undef	SKIPN
#undef	C2H
}

/*
 * The inverse of send_state() -- take input in hex and convert it into
 * a Mach saved_state structure.  Return a new pointer to the input buffer.
 */
char *
receive_state(cp, ssp)
	register char *cp;
	register struct hp800_saved_state *ssp;
{
#define	H2M(m, l)	cp = hex2mem(cp, (char *) &ssp->m, l * sizeof (int))
#define	SKIPN(n)	cp += n * 2 * sizeof (int)

	H2M(flags, 1);
	H2M(r1, 31);
	H2M(cr11, 1);
	H2M(iioq_head, 1);
	H2M(iisq_head, 1);
	H2M(iioq_tail, 1);
	H2M(iisq_tail, 1);
	H2M(cr15, 1);
	H2M(cr19, 4);
	SKIPN(1);
	H2M(sr4, 1);
	H2M(sr0, 4);
	H2M(sr5, 6);
	SKIPN(1);
	H2M(cr12, 2);
	SKIPN(7);
	H2M(fr0, 32);

#undef	H2M
#undef	SKIPN

	return cp;
}

void
kgdb_fixpc(type, ssp, pcb)
	int type;
	struct hp800_saved_state *ssp;
	pcb_t pcb;
{
	unsigned insn;
	extern void user_nullify();

	if (pcb->kgdb_trace & KGDB_STEP) {
		/*
		 * If we performed a non-nullified branch-and-link,
		 * the link register will point to user_nullify;
		 * we must make it point after the original tail.
		 */
		if (type == I_TAKEN_BR) {
			insn = pcb->saved_insn;
			if ((insn & 0xfc000000) == 0xe4000000)	/* ble */
				ssp->r31 = pcb->saved_oqt + 4;
			else
			if ((insn & 0xfc000000) == 0xe8000000 &&
			    ((insn & 0xe000) == 0x4000 ||	/* blr */
			     (insn & 0xe000) == 0) &&		/* bl */
			    (insn & 0x3e00000) != 0) {		/* link != 0 */
				((int *)ssp)[(insn & 0x3e00000) >> 21] =
					pcb->saved_oqt + 4;
			}
			/* XXX also fiddle tail space on BE? */
		}

		/*
		 * If we got to the break at user_nullify,
		 * we must have executed an instruction that
		 * nullified the user_sstep break but did not branch.
		 * In this case we must restore the PSW N bit that was
		 * cleared at the successfully executed break.
		 */
		if (ssp->iioq_head == user_nullify)
			ssp->ipsw |= PSW_N;

		/*
		 * If we didn't branch (the head or tail
		 * of the queue points after user_sstep),
		 * then the new tail follows the new head.
		 */
		if (ssp->iioq_head == user_nullify ||
		    ssp->iioq_tail == user_nullify) {
			ssp->iisq_tail = pcb->saved_sqt;
			ssp->iioq_tail = pcb->saved_oqt + 4;
		}

		/*
		 * Restore the original succeeding pc.
		 */
		ssp->iisq_head = pcb->saved_sqt;
		ssp->iioq_head = pcb->saved_oqt;

		ssp->ipsw &= ~PSW_T;
		pcb->kgdb_trace &= ~KGDB_STEP;
	}

	/*
	 * If we arrived at kgdb_break, force us to continue
	 * past this point.
	 */
	if (ssp->iioq_head == kgdb_break) {
		ssp->iioq_head += 4;
		if (ssp->iioq_head == ssp->iioq_tail)
			ssp->iioq_tail += 4;
	}
}

/*
 * Process threads that request a kgdb action.
 * We politely ask the kgdb thread to switch to our context.
 */
int 
kgdb_trap(type, ssp)
	int type;
	struct hp800_saved_state *ssp;
{
	int s;
	thread_t t = current_thread();

	if (istackptr == 0)
		/* on the interrupt stack */
		return kgdb_interrupt(type, ssp);

	if (kgdb_dev < 0 || t == 0 || t->pcb == 0)
		/* not debugging */
		return (0);

	s = splvm();	/* block other kgdb traps */

	/*
	 * Try to get the kgdb thread's attention.
	 */
	t->pcb->kgdb_trace |= KGDB_INPROGRESS;
	do {
		if (kgdb_lock) {
			if (t->state & TH_IDLE) {
				/*
				 * The idle thread can't wait.
				 */
				t->pcb->kgdb_trace &= ~KGDB_INPROGRESS;
				break;
			}
			printf("kgdb: context %x blocking\n", t);
			assert_wait(&kgdb_lock, 1);
			thread_block((void (*)()) 0);
			continue;	/* note that kgdb_trace may change */
		}
		kgdb_lock = 1;
		kgdb_context = t;
		kgdb_ssp = ssp;
		kgdb_type = type;
		kgdb_spl = s;
		thread_wakeup(&kgdb_context);
		if (!(t->state & TH_IDLE))
			assert_wait(&kgdb_lock, 1);
		thread_block((void (*)()) 0);
	} while (t->pcb->kgdb_trace & KGDB_INPROGRESS);

	splx(s);
	return 1;
}

/*
 * We come here for a single-threaded kgdb when executing
 * from the interrupt stack.
 */
int
kgdb_interrupt(type, ssp)
	int type;
	struct hp800_saved_state *ssp;
{
	int s;
	static struct pcb ipcb;		/* to hold debugging state */
	extern vm_map_t kernel_map;

	if (kgdb_dev < 0)
		/* not debugging */
		return (0);

	s = splvm();			/* we can still take hardclocks */

	if (kgdb_lock) {
		/*
		 * Can't afford to wait when we're on the interrupt stack.
		 */
		panic("breakpoint in interrupt can't be serviced"); /*XXX*/
		splx(s);
		return 0;
	}
	kgdb_lock = 1;

	kgdb_fixpc(type, ssp, &ipcb);
	(void) kgdb_command(type, ssp, &ipcb, kernel_map);

	kgdb_lock = 0;
	splx(s);
}

int kgdb_dummy;

/*
 * The dummy thread exists merely as a context in which to take breakpoints.
 * The serial port driver wakes up the dummy thread when
 * a kgdb attention character ('!') is received.
 */
void
kgdb_dummy_thread()
{
	extern void kgdb_break();

	while (1) {
		assert_wait(&kgdb_dummy, 1);
		thread_block((void (*)()) 0);
		kgdb_break();
	}
}

int kgdb_xpackets = 0;

void
sendxpacket(context)
	thread_t context;
{
	task_t pid = context->task;
	extern pmap_t kernel_pmap;

	if (pid->map->pmap == kernel_pmap) {
		if (kgdb_xpackets == 0)
			return;
		pid = 0;
	}

	kgdb_xpackets = 1;

	outbuffer[0] = 'X';
	(void)val2hex(&outbuffer[1], (int) pid, 8);
	putpacket(outbuffer);
}

/*
 * A kernel thread that handles other threads that need kgdb service.
 * For each context, we run the kgdb command loop.
 * Requests to switch contexts can come from kgdb_trap() or from commands.
 */
kgdb_thread()
{
	thread_t old_kgdb_context = (thread_t) -1;
	int needxpacket = 0;
	extern vm_map_t kernel_map;
	vm_map_t map;

	current_thread()->max_priority = BASEPRI_SYSTEM - 1;
	current_thread()->priority = BASEPRI_SYSTEM - 1;
	current_thread()->sched_pri = BASEPRI_SYSTEM - 1;

	printf("kgdb_thread startup\n");

	while (1) {
		/*
		 * If we don't have a context, go out and get one.
		 */
		(void) splvm();
		if (kgdb_context == (thread_t) -1) {
			kgdb_lock = 0;
			thread_wakeup(&kgdb_lock);
			assert_wait(&kgdb_context, 1);
			thread_block((void (*)()) 0);
			if (kgdb_context == (thread_t) -1)
				continue;
			if (old_kgdb_context != kgdb_context)
				needxpacket = 1;
		}
		splx(kgdb_spl);

		if (old_kgdb_context != kgdb_context) {
#if 0
			printf("kgdb: debugging context %x\n",
				kgdb_context);
#endif
			/* clear_kgdb_page_cache(); */
		}
		old_kgdb_context = kgdb_context;

		/*
		 * Recover from any pc queue manipulations that we may
		 * have made in order to accomplish a single step.
		 */
		kgdb_fixpc(kgdb_type, kgdb_ssp, kgdb_context->pcb);

		/*
		 * If we arrived here on a breakpoint and
		 * our pc was in the kernel, use the kernel map;
		 * otherwise use the current task's map.
		 * This allows us to debug in the kernel
		 * on behalf of user threads.
		 * If we need an X packet, send one that's appropriate
		 * to the map we chose.
		 */
		if ((kgdb_type == I_BRK_INST || kgdb_type == I_TAKEN_BR) &&
		    kgdb_ssp->iisq_head == HP800_SID_KERNEL) {
			map = kernel_map;
			if (needxpacket)
				sendxpacket(current_thread());
		} else {
			map = kgdb_context->task->map;
			if (needxpacket)
				sendxpacket(kgdb_context);
		}
		needxpacket = 0;

		/*
		 * Run the command loop in the new context.
		 */
printf("kgdb_thread: kgdb_context = %x\n", kgdb_context);
		kgdb_context = kgdb_command(kgdb_type, kgdb_ssp,
					    kgdb_context->pcb, map);
	}
}

#define RESPOND(str) (bcopy(str, outbuffer, sizeof(str)))

/*
 * This function does all command processing for interfacing to 
 * a remote gdb.
 */
thread_t
kgdb_command(type, ssp, pcb, map)
	int type;
	struct hp800_saved_state *ssp;
	pcb_t pcb;
	vm_map_t map;
{
	int sigval;
	int addr, length;
	char *ptr;
	extern void user_sstep();
	extern int istackptr;
	extern task_t kernel_task;
	extern pmap_t kernel_pmap;

        if (kgdb_debug)
		printf("type=%d, pc=%x/%x, psw=0x%x, pcb=%x, map=%x\n",
			type, ssp->iisq_head, ssp->iioq_head, ssp->ipsw, pcb, map);

	if (map->pmap != kernel_pmap)
		kgdb_xpackets = 1;

	/*
	 * If the user sets this, switch contexts.
	 */
	kgdb_context = (thread_t) -1;

	/*
	 * Get a Unix signal to make GDB happy.
	 */
	sigval = computeSignal(type);

	/* reply to host that an exception has occurred */
	outbuffer[0] = 'S';
	(void)val2hex(&outbuffer[1], sigval, 2);
	putpacket(outbuffer);

	while (kgdb_context == (thread_t) -1) {
		outbuffer[0] = 0;
		getpacket(inbuffer);
		ptr = inbuffer;
		switch (*ptr++) {
		case '?':
			outbuffer[0] = 'S';
			(void)val2hex(&outbuffer[1], sigval, 2);
			break;

		case 'g':	/* return the value of the CPU registers */
			ptr = outbuffer;
			ptr = mem2hex(ptr, (char *) ssp,
#ifdef hp800
				      sizeof(*ssp) - 4);
#else
				      sizeof(*ssp) - 4 - (16*sizeof(double)));
#endif
			break;

		case 'G':	/* set the value of the CPU registers */
			ptr = hex2mem(ptr, (char *) ssp,
#ifdef hp800
				      sizeof(*ssp) - 4);
#else
				      sizeof(*ssp) - 4 - (16*sizeof(double)));
#endif
			RESPOND("OK");
			break;

			/* mAA..AA,LLLL  Read LLLL bytes at address AA..AA */
		case 'm':
			ptr = hex2val(ptr, &addr, 8);
			if (*ptr++ != ',') {
				RESPOND("E01");
				debug_error("malformed read memory cmd: %s\n",
					    inbuffer);
				break;
			}
			ptr = hex2val(ptr, &length, 8);
			if (length <= 0 || length >= BUFMAX*2) {
				RESPOND("E02");
				if (kgdb_debug)
					printf("bad read memory length: %d\n",
					       length);
				break;
			}
			if (! kgdbacc(map, &addr, length, VM_PROT_READ)) {
				RESPOND("E03");
				if (kgdb_debug)
					printf("read access violation addr 0x%x len %d\n", addr, length);
				break;
			}
			(void)mem2hex(outbuffer, (char *)addr, length);
			break;

			/*
			 * MAA..AA,LLLL: Write LLLL bytes at address AA.AA
			 * return OK
			 */
		case 'M':
			/*
			 * XXX  If we modify a pdir entry,
			 * we should flush the tlb
			 * for the corresponding page(s)...
			 */
			ptr = hex2val(ptr, &addr, 8);
			if (*ptr++ != ',') {
				RESPOND("E01");
				debug_error("malformed write memory cmd: %s\n",
					    inbuffer);
				break;
			}
			ptr = hex2val(ptr, &length, 8);
			if (*ptr++ != ':') {
				RESPOND("E01");
				debug_error("malformed write memory cmd: %s\n",
					    inbuffer);
				break;
			}
			if (length <= 0 || length >= BUFMAX*2 - 32) {
				RESPOND("E02");
				if (kgdb_debug)
					printf("bad write memory length: %d\n",
					       length);
				break;
			}
			if (! kgdbacc(map, &addr, length, VM_PROT_READ|VM_PROT_WRITE)) {
				RESPOND("E03");
				if (kgdb_debug)
					printf("write access violation addr 0x%x len %d\n", addr, length);
				break;
			}
			(void)hex2mem(ptr, (char *)addr, length);
			RESPOND("OK");
			break;

			/*
			 * cAA..AA  Continue at address AA..AA
			 * (address optional)
			 */
		case 'c':
			/*
			 * try to read optional start address.
			 */
			if (ptr != hex2val(ptr, &addr, 8)) {
				ssp->iioq_head = addr;
				ssp->iioq_tail = addr + 4;
				if (kgdb_debug)
					printf("new pc = 0x%x\n", addr);
			}

			ssp->ipsw &= ~PSW_T;
			pcb->kgdb_trace &= ~KGDB_INPROGRESS;
			return (thread_t) -1;

			/*
			 * sAA..AA  Step one instruction from AA..AA
			 * (address optional)
			 */
		case 's':
			/*
			 * try to read optional start address.
			 */
			if (ptr != hex2val(ptr, &addr, 8)) {
				ssp->iioq_head = addr;
				ssp->iioq_tail = addr + 4;
				if (kgdb_debug)
					printf("new pc = 0x%x\n", addr);
			}

			addr = ssp->iioq_head;
			if (!kgdbacc(map, &addr, sizeof (long), VM_PROT_READ)) {
				RESPOND("E03");
				if (kgdb_debug)
					printf("read access violation addr 0x%x len %d\n", addr, length);
				break;
			}

			/*
			 * Put a break instruction in the pc queue so that
			 * after executing the queue head, we trap.
			 */
			pcb->saved_oqt = ssp->iioq_tail;
			pcb->saved_sqt = ssp->iisq_tail;
			ssp->iioq_tail = (unsigned) &user_sstep;
			ssp->iisq_tail = HP800_SID_KERNEL;

			/*
			 * Set the taken branch trap bit in the PSW so that
			 * we can't branch around the breakpoint.
			 * Save the instruction at the head of the queue
			 * in case we need to fix a link register after
			 * a taken branch trap on a bl or blr.
			 */
			ssp->ipsw |= PSW_T;
#define	pcalign(pc)	(int *)((int)(pc) & ~3)
			pcb->saved_insn = *pcalign(addr);

			pcb->kgdb_trace |= KGDB_STEP;
			pcb->kgdb_trace &= ~KGDB_INPROGRESS;
			return (thread_t) -1;

			/* kill the program (same as continue for now) */
		case 'k':
			ssp->ipsw &= ~PSW_T;
			pcb->kgdb_trace &= ~KGDB_INPROGRESS;
			return (thread_t) -1;

		case 'X':
			/*
			 * Switch to a new context.
			 * In Unix, this would be a pid;
			 * in Mach, it's a thread pointer.
			 */
			(void) hex2val(ptr, &kgdb_context, 8);
			kgdb_xpackets = 1;
			continue;
		}
		/* reply to the request */
		putpacket(outbuffer);
	}

	if (istackptr == 0)
		/*
		 * Sorry, we're on the interrupt stack.
		 */
		return (thread_t) -1;
	if (kgdb_context == 0)
		kgdb_context = (thread_t) kernel_task;
	if (is_a_task((task_t) kgdb_context))
		kgdb_context = kgdb_pick_thread((task_t) kgdb_context);
	else if (!is_a_thread(kgdb_context)) {
		printf("kgdb: %x: not a thread\n", addr);
		return (thread_t) -1;
	}
	if (inbuffer[0] != 'X')
		sendxpacket(kgdb_context);
	kgdb_ssp = (struct hp800_saved_state *)(kgdb_context->pcb + 1);
	kgdb_type = -SIGINT;
	kgdb_spl = SPL0;
	return kgdb_context;
}

int
is_a_task(addr)
	vm_offset_t addr;
{
	register task_t		task;
	processor_set_t		pset;

	simple_lock(&all_psets_lock);
	pset = (processor_set_t) queue_first(&all_psets);
	while (!queue_end(&all_psets, (queue_entry_t) pset)) {
		pset_lock(pset);
		task = (task_t) queue_first(&pset->tasks);
		while (!queue_end(&pset->tasks, (queue_entry_t) task)) {
			if (addr == (vm_offset_t) task) {
				pset_unlock(pset);
				simple_unlock(&all_psets_lock);
				return 1;
			}
			task = (task_t) queue_next(&task->pset_tasks);
		}
		pset_unlock(pset);
		pset = (processor_set_t) queue_next(&pset->all_psets);
	}
	simple_unlock(&all_psets_lock);

	return 0;
}

int
is_a_thread(addr)
	vm_offset_t addr;
{
	register thread_t	thread;
	processor_set_t		pset;

	simple_lock(&all_psets_lock);
	pset = (processor_set_t) queue_first(&all_psets);
	while (!queue_end(&all_psets, (queue_entry_t) pset)) {
		pset_lock(pset);
		thread = (thread_t) queue_first(&pset->threads);
		while (!queue_end(&pset->threads, (queue_entry_t) thread)) {
			if (addr == (vm_offset_t) thread) {
				pset_unlock(pset);
				simple_unlock(&all_psets_lock);
				return 1;
			}
			thread = (thread_t) queue_next(&thread->pset_threads);
		}
		pset_unlock(pset);
		pset = (processor_set_t) queue_next(&pset->all_psets);
	}
	simple_unlock(&all_psets_lock);

	return 0;
}

/*
 * Pick a thread to debug out of a task's thread lists.
 * If we find an INPROGRESS thread, we use it; otherwise it's the first thread.
 * When we pick a thread, we move it to the front of the list.
 */
thread_t
kgdb_pick_thread(task)
	task_t task;
{
	register queue_head_t *list;
	register thread_t thread;

	task_lock(task);
	list = &task->thread_list;
	thread = (thread_t) queue_first(list);
	while (!queue_end(list, (queue_entry_t) thread)) {
		if (thread->pcb->kgdb_trace & KGDB_INPROGRESS) {
			queue_remove(list, thread, thread_t, thread_list);
			queue_enter_first(list, thread, thread_t, thread_list);
			task_unlock(task);
			return thread;
		}
		thread = (thread_t) queue_next(&thread->thread_list);
	}
	task_unlock(task);
	return (thread_t) queue_first(list);
}

vm_map_t cached_map = 0;
vm_prot_t cached_protection = VM_PROT_NONE;
vm_offset_t cached_pages = -1;
vm_offset_t kernel_pages = -1;
int cache_size = 0;

void
clear_kgdb_page_cache() {
	extern vm_map_t kernel_map;

	if (cached_map == 0)
		return;
	cached_map = 0;
	cached_protection = VM_PROT_NONE;
	cached_pages = -1;
	if (kernel_pages != -1) {
		/*
		 * vm_map_remove() releases the page(s) and
		 * decrements the object's reference count.
		 */
		if (vm_map_remove(kernel_map, kernel_pages,
				  kernel_pages + cache_size)
		    != KERN_SUCCESS)
			printf("kgdb: failed to remove address %x, length %x\n",
				kernel_pages, cache_size);
	}
	kernel_pages = -1;
	cache_size = 0;
}

/*
 * Make an address from a particular map available if we can.
 * If the page isn't resident, try to fault it in.
 * If we map the data to some other address, change *addrp appropriately.
 */
int
kgdbacc(map, addrp, count, rw)
	vm_map_t map;
	vm_offset_t *addrp;
	vm_size_t count;
	vm_prot_t rw;
{
	vm_map_version_t version;
	vm_object_t object;
	vm_offset_t offset;
	vm_prot_t prot;
	boolean_t wired;
	boolean_t single_use;
	vm_size_t size;
	vm_prot_t max_prot;
	vm_inherit_t inheritance;
	boolean_t is_shared;
	mach_port_t object_name;
	vm_offset_t offset_in_object;
	vm_offset_t va;
	vm_offset_t vbegin, vend;
	register int i;
	int s;

	if (map->pmap == kernel_pmap)
		return kernacc(*addrp, count, rw);

	vbegin = trunc_page(*addrp);
	vend = round_page(*addrp + count);

	/*
	 * Look for a cache hit.
	 */
	if (map == cached_map &&
	    (rw & cached_protection) == rw &&
	    vbegin >= cached_pages &&
	    vend <= cached_pages + cache_size) {
		*addrp = kernel_pages + (*addrp - vbegin);
		return 1;
	}

	clear_kgdb_page_cache();

	/*
	 * Find the first page in the target map.
	 * Note that the object is returned locked, but not referenced.
	 */
	if (vm_map_lookup(&map, vbegin, VM_PROT_READ, &version, &object,
			  &offset, &prot, &wired, &single_use) !=
	    KERN_SUCCESS)
		return 0;

#if 0
	if ((rw & prot) != rw) {
		/*
		 * We may have to modify the page's protections.
		 * Get information about the current protection.
		 */
		va = vbegin;
		if (vm_region(map, &va, &size, &prot, &max_prot,
			      &inheritance, &is_shared, &object_name,
			      &offset_in_object) != KERN_SUCCESS) {
			vm_object_unlock(object);
			return 0;
		}

		/*
		 * If the page is read-only and not COW,
		 * make it COW so that tasks that share it
		 * don't see our modifications.
		 */
		if (inheritance == VM_INHERIT_SHARE)
			(void) vm_map_inherit(map, vbegin, vend,
					      VM_INHERIT_COPY);
	}
#endif

	/*
	 * Put the page in the kernel map.
	 */
	vm_object_reference(object);
	va = 0;
	if (vm_map_enter(kernel_map, &va, vend - vbegin,
			 (vm_offset_t) 0, TRUE, object, offset,
			 FALSE, rw, rw, VM_INHERIT_NONE) !=
	    KERN_SUCCESS) {
		vm_object_unlock(object);
		vm_object_deallocate(object);
		printf("kgdb: couldn't map address %x from map %x\n",
		       vbegin, map);
		return 0;
	}
	vm_object_unlock(object);

	/*
	 * If the page isn't in memory and wired down, get it.
	 */
	if (!wired) {
		/*
		 * Fault the page in.
		 * XXX what if IPL is so high, it prevents page-ins?
		 */
		s = splsched();
	    	thread_lock(current_thread());
		i = current_thread()->state & TH_UNINT;
		current_thread()->state |= TH_UNINT;
		thread_unlock(thread);
		splx(s);
		if (vm_map_pageable(kernel_map, va,
				    va + (vend - vbegin), rw) !=
		    KERN_SUCCESS) {
			if (i == 0) {
				s = splsched();
	    			thread_lock(current_thread());
				current_thread()->state &= ~TH_UNINT;
				thread_unlock(thread);
				splx(s);
			}
			printf("kgdb: couldn't fault in address %x from map %x\n",
			       va, map);
			return 0;
		}
		if (i == 0) {
			s = splsched();
	    		thread_lock(current_thread());
			current_thread()->state &= ~TH_UNINT;
			thread_unlock(thread);
			splx(s);
		}
	}

	kernel_pages = va;
	cached_map = map;
	cached_protection = rw;
	cached_pages = vbegin;
	cache_size = vend - vbegin;

	*addrp = kernel_pages + (*addrp - vbegin);
	return 1;
}

/* XXX WE CURRENTLY MAP ALL MEMORY AS WRITABLE */
/* XXX THESE ROUTINES COMPLETELY IGNORE MACH PROTECTIONS */
/* XXX THESE ROUTINES WON'T ALLOW GDB TO ACCESS DYNAMIC KERNEL MEMORY */

/*
 * See if count bytes starting at VA addr are accessible.
 * At the moment kernel text is writable, so we ignore protections.
 * We don't do any sort of locking here; could be dangerous occasionally...
 */
kernacc(addr, count, rw)
	unsigned int addr;
	int count, rw;
{
	register vm_offset_t	va;
	register vm_offset_t	vend = round_page(addr + count);

	/* XXX ugly hack for negative numbers masquerading as addresses... */
	if ((addr & 0xFFF80000) == 0xFFF80000)
		return 0;

	for (va = trunc_page(addr); va != vend; va += HP800_PGBYTES)
		if (pmap_extract(kernel_pmap, va) == 0)
			return 0;
	return 1;
}

/*
 * Change protections on kernel pages from addr to addr+size
 * (presumably so debugger can plant a breakpoint).
 * All addresses are assumed to reside in the Sysmap,
 */
void
chgkprot(addr, size, rw)
	register vm_offset_t addr;
	int size, rw;
{
	/* XXX IGNORE FOR NOW */
}

#endif /* KGDB */
